#!/usr/bin/octave -qf

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.

function make_Gibson_pres()
     filename = "xy/gibson-pres.xy";
     out = fopen(filename, 'wt');

     sec_in_Ma = 3.15e+13;

     grav = 10.0;
     rhob = 2100.0;
     rhof = 1000.0;

     tmax = 1 * sec_in_Ma;
     k = 1.0e-18;
     alpha = 1.0e-8;
     mu = 1.0e-3;
     hmax = 1000.0;

     cv = k / (alpha * mu);
     m = hmax / tmax;
     gamma_prime = (rhob - rhof) * grav;

     fprintf(out, "#( (permeability) k = %g [m2] )#\n", k);
     fprintf(out, "#( (compressibility) alpha = %g [1/Pa] )#\n", alpha);
     fprintf(out, "#( (viscosity) mu = %g [Pas] )#\n", mu);
     fprintf(out, "#( cv = %g [m2/s] )#\n", cv);
     fprintf(out, "#( m = %g [m/Ma] )#\n", m);
     fprintf(out, "#( gamma_prime = %g [Pa/m] )#\n", gamma_prime);
     fprintf(out, "#( height:0 depth:1 pres:2  )#\n");
     fprintf(out, "#( [m]:0    [m]:1   [MPa]:2 )#\n");

     n_time_steps = 5;
     n_cells = 50;

     dt = tmax / n_time_steps;
     dx = hmax / n_cells;

     t = dt;
     while (t < tmax + 0.1)
          h = m * t;
          t_ma = t / sec_in_Ma;

          printf("(t=%g Ma) ", t_ma);
          fprintf(out, "\n");
          fprintf(out, "#( t=%g Ma )#\n", t_ma);

          x = 0;
          while (x < h + 0.1)
               pres = get_gibson_pres(gamma_prime, cv, m, x, t);
               fprintf(out, "%10.4f %10.4f %10.4f\n", x, -(h - x), (pres / 1e+6));
               fflush(out);
               x += dx;
          end

          if (h > 1.0)
               p1 = (rhob - rhof) * grav * h;
               p2 = get_gibson_pres(gamma_prime, cv, m, 0.0, t);
               p3 = p2 / p1;
               printf("(p_max=%g Pa) (p_ex_lith=%g Pa) (p_hat=%g)\n", p2, p1, p3);
          endif

          t += dt;
     end

     printf("(results are written to: %s)\n", filename);
endfunction

function pres = get_gibson_pres(gamma_prime, cv, m, x, t)
     part1 = gamma_prime * m * t;
     fac1 = gamma_prime * exp( - x * x / (4.0 * cv * t)) / sqrt(pi * cv * t);
     fac2 = get_gibson_integral(cv, m, x, t);
     part2 = fac1 * fac2;
     pres = part1 - part2;
endfunction

function I = get_gibson_integral(cv, m, x, t)
     dzeta = 1;
     I = 0.0;

     zeta = 0.0;
     is_finished = 0;
     while (! is_finished)
          fac1 = tanh(m * zeta / (2.0 * cv));
          fac2 = cosh(x * zeta / (2.0 * cv * t));
          fac3 = exp(- zeta * zeta / (4.0 * cv * t));

          dI = zeta * fac1 * fac2 * fac3 * dzeta;

          if ((dI < 1.0e-18 * I) || (zeta > 1e+35))
               is_finished = 1;
          endif

          I += dI;
          zeta += dzeta;
          dzeta *= 1.0001;
     end
endfunction

make_Gibson_pres();
