/*
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; version 2 of the License.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   (c) Copyright 1998-2009 by M. Wangen, IFE, Box 40, N-2007 Kjeller, Norway.
**
**   File: flexure.c
**   Title: Compute the deflection of the lithosphere due to a load.
**   Date: Version 1.1, April 2002
**   Author: Magnus Wangen, IFE, magnus@ife.no
**
**   $Id$
*/

/*
**   History:
**   Version 1.0, Apr 1998: First operative version for UNIK use.
**   Version 1.1, Apr 2002: Added checking of the input data.
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef M_PI
#     define M_PI 3.14159
#endif

#define MODE_ELASTIC 1
#define MODE_VISCOELASTIC 2
#define MODE_VISCOUS 3

#define MAX_OUT 1024
#define MAX_LOAD 1024
#define pi 3.14159265358979323846
#define MATCH(a, b) (strcmp(a,b) == 0)
#define ERROR_EXIT(text) {fprintf(stderr, "%s\n", text); exit(1);}
#define ABS(x) ((x) > 0.0 ? (x) : (-(x)))
#define SEC_IN_MA 3.1536e+13

int main(int argc, char **argv);
void read_options(int argc, char **argv);
void show_options(FILE *out);
void check_input_data(void);
void make_derived_params(void);
void write_elastic_deflection(void);
void write_viscoelastic_deflection(void);
void write_viscous_deflection(void);
void write_params(FILE *out);
double get_rectangle_fourier_part(int n, int k);
double get_rectangle_dI(int n, double x, double x1, double x2, double h1, double h2);
double get_point_part(int k, double x);
double get_point_load_deflection(double x, double xm, double V0);
double get_one_point_load(double x);
double get_total_area(void);


const char 
     *filename = "dataRectangleFourier.xy";

static double
     xload[MAX_LOAD],
     hload[MAX_LOAD],
     t_hat_out[MAX_OUT],
     F_factor = 1.0,
     L = 1.0,
     E = 70.0e+9,
     Te = 30.0e+3,
     nu = 0.25,
     mu = 1.0e+24,
     rhom = 3300.0,
     rhoc = 2700.0,
     grav = 10.0;

static double
     D = 0.0,
     Dv = 0.0,
     F0 = 0.0,
     drho_g = 0.0,
     alpha = 0.0;

static int
     mode = MODE_ELASTIC,
     debug_wave_lengths = 0,
     t_hat_out_size = 0,
     load_size = 0,
     grid_size = 100,
     infinity = 100;


int main(
     int argc, 
     char **argv)
{
     read_options(argc, argv);
     check_input_data();
     make_derived_params();
     write_params(stdout);

     if (mode == MODE_ELASTIC)
          write_elastic_deflection();
     else if (mode == MODE_VISCOELASTIC)
          write_viscoelastic_deflection();
     else if (mode == MODE_VISCOUS)
          write_viscous_deflection();
     else
          ERROR_EXIT("Illegal mode!");

     return 0;
}


void read_options(
     int argc,
     char **argv)
{
     int i, k;
     int got_errors = 0;

     for (i = 1; i < argc; i++)
          if      (MATCH(argv[i], "-inf")) infinity = atoi(argv[++i]);
          else if (MATCH(argv[i], "-grid")) grid_size = atoi(argv[++i]);
          else if (MATCH(argv[i], "-L")) L = atof(argv[++i]);
          else if (MATCH(argv[i], "-E")) E = atof(argv[++i]);
          else if (MATCH(argv[i], "-nu")) nu = atof(argv[++i]);
          else if (MATCH(argv[i], "-mu")) mu = atof(argv[++i]);
          else if (MATCH(argv[i], "-Te")) Te = atof(argv[++i]);
          else if (MATCH(argv[i], "-Ffac")) F_factor = atof(argv[++i]);
          else if (MATCH(argv[i], "-file")) filename = argv[++i];
          else if (MATCH(argv[i], "-elastic")) mode = MODE_ELASTIC;
          else if (MATCH(argv[i], "-viscoelastic")) mode = MODE_VISCOELASTIC;
          else if (MATCH(argv[i], "-viscous")) mode = MODE_VISCOUS;
          else if (MATCH(argv[i], "-showlambda")) debug_wave_lengths = 1;
          else if (MATCH(argv[i], "-help") || MATCH(argv[i], "-h"))
          {
               show_options(stdout);
               exit(1);
          }
          else if (MATCH(argv[i], "-xh"))
          {
               if (load_size >= MAX_LOAD)
                    ERROR_EXIT("Too many load positions!");

               xload[load_size] = atof(argv[++i]);
               hload[load_size] = atof(argv[++i]);
               load_size++;
          }
          else if (MATCH(argv[i], "-show"))
          {
               for (k = 0; k < load_size; k++)
                    printf(" -xh %g %g", xload[k], hload[k]);

               printf(" -L %g\n", L);
          }
          else if (MATCH(argv[i], "-case1")) 
          {
               L = 1000e+3;
               xload[0] = 0.40 * L;
               xload[1] = 0.45 * L;
               xload[2] = 0.55 * L;
               xload[3] = 0.60 * L;
               hload[0] = 0.0;
               hload[1] = 2000.0;
               hload[2] = 2000.0;
               hload[3] = 0.0;
               load_size = 4;
          }
          else if (MATCH(argv[i], "-tHat"))
          {
               if (t_hat_out_size >= MAX_OUT)
                    ERROR_EXIT("Too many output steps!");

               t_hat_out[t_hat_out_size] = atof(argv[++i]);
               t_hat_out_size++;
          }
          else
          {
               fprintf(stderr, "unknown option: %s\n", argv[i]);
               got_errors++;
          }

     if (got_errors)
     {
          fprintf(stderr, "Got %d errors! Use -help to get help!\n", got_errors);
          exit(1);
     }
}


void show_options(
     FILE *out)
{
     fprintf(out, "flexure <options>\n");
     fprintf(out, "     -help -h            Show this help.\n");
     fprintf(out, "     -inf <numb>         Set number of terms in the Fourier series.\n");
     fprintf(out, "     -grid <numb>        Set number of points in the grid.\n");
     fprintf(out, "     -L <numb>           Set grid size (length), [m].\n");
     fprintf(out, "     -E <numb>           Set Young's modulus, [Pa].\n");
     fprintf(out, "     -nu <numb>          Set Poisson ratio.\n");
     fprintf(out, "     -mu <numb>          Set viscosity, [Pas].\n");
     fprintf(out, "     -Te <numb>          Set the effective elastic thickness, [m].\n");
     fprintf(out, "     -Ffac <numb>        Set a factor to be multiplied with F, [-].\n");
     fprintf(out, "     -xh <numb> <numb>   Set load x-pos and height, [m].\n");
     fprintf(out, "     -show               Show the options setting the load.\n");
     fprintf(out, "     -showlambda         Show stable and unstable wave lengths.\n");
     fprintf(out, "     -case1              Use a demo case.\n");
     fprintf(out, "     -elastic            Elastic (time-independent) flexure,\n");
     fprintf(out, "     -viscoelastic       Viscoelastic (time-dependent) flexure,\n");
     fprintf(out, "     -viscous            Viscous (time-dependent) flexure,\n");
     fprintf(out, "     -file <name>        Set output filename.\n");
     fprintf(out, "use: -case1 -show to see the input for the demo case.\n");
}


void check_input_data(
     void)
{
     int i;

     if (load_size < 3)
          ERROR_EXIT("Need at least 3 load points!");

     for (i = 1; i < load_size; i++)
          if (xload[i] < xload[i - 1])
               ERROR_EXIT("The x-coordinates must by in increasing order!");

     if (xload[0] < 0.0)
          ERROR_EXIT("The first x-coordinate must be large than 0!");

     if (xload[load_size - 1] > L)
          ERROR_EXIT("The last x-coordinate must be less than the grid length!");

     if (ABS(hload[0]) > 1.0e-6)
          ERROR_EXIT("The first load height must be 0!");
          
     if (ABS(hload[load_size - 1]) > 1.0e-6)
          ERROR_EXIT("The last load height must be 0!");
}


void make_derived_params(
     void)
{
     double term;

     D = E * Te * Te * Te / (12.0 * (1 - nu));
     Dv = mu * Te * Te * Te / 12.0;
     drho_g = (rhom - rhoc) * grav;
     F0 = 0.5 * rhoc * grav * Te * Te;
     
     term = 4.0 * D / drho_g;
     alpha = exp(log(term) / 4.0);
}


void write_elastic_deflection(
     void)
{
     int k, n;
     FILE *out = fopen(filename, "w");
     double dx, x, zload, wfourier, wpoint, wone, an, bn, kn, kn2, kn4;

     write_params(out);
     fprintf(out, "#( x:0 zLoad:1 wFourier:2 wPoint:3 wOne:4 x:5 wFourier:6 )#\n");

     dx = L / grid_size;

     for (x = 0.0; x < L; x += dx)
     {
          zload = 0.0;
          wfourier = 0.0;
          wpoint = 0.0;

          for (n = 1; n < infinity; n++)
          {
               an = 0.0;

               for (k = 0; k < load_size - 1; k++)
                    an += get_rectangle_fourier_part(n, k);

               kn = n * pi / L;
               kn2 = kn * kn;
               kn4 = kn2 * kn2;

               zload += an * sin(kn * x);

               bn = an * rhoc * grav / (D * kn4 + drho_g);
               wfourier -= bn * sin(kn * x);
          }

          for (k = 0; k < load_size - 1; k++)
               wpoint -= get_point_part(k, x);

          wone = - get_one_point_load(x);

          fprintf(out, "%g %g %g %g %g %g %g\n", 
               x, zload, wfourier, wpoint, wone,
               (x / 1000.0), (wfourier / 1000.0));
     }

     fclose(out);
     printf("(data for elastic deflection is written to: %s)\n", filename);
}


void write_viscoelastic_deflection(
     void)
{
     int i, k, n;
     FILE *out = fopen(filename, "w");
     double dx, x, zload, wfourier, an, bn, kn, kn2, kn4;
     double t, te, t0, Yt;

     write_params(out);

     te = mu / E;

     if (t_hat_out_size <= 0)
     {
          t_hat_out[0] = 0.0;
          t_hat_out[1] = 1.0;
          t_hat_out[2] = 1000.0;
          t_hat_out_size = 3;
     }

     dx = L / grid_size;

     for (i = 0; i < t_hat_out_size; i++)
     {
          t = t_hat_out[i] * te;

          fprintf(out, "\n");
          fprintf(out, "#( t=%g [Ma], tHat=%g )#\n", (t / SEC_IN_MA), (t / te));
          fprintf(out, "#( x:0 zLoad:1 wFourier:2 x:3 wFourier:4)#\n");

          for (x = 0.0; x < L; x += dx)
          {
               zload = 0.0;
               wfourier = 0.0;
     
               for (n = 1; n < infinity; n++)
               {
                    an = 0.0;
     
                    for (k = 0; k < load_size - 1; k++)
                         an += get_rectangle_fourier_part(n, k);
     
                    kn = n * pi / L;
                    kn2 = kn * kn;
                    kn4 = kn2 * kn2;
     
                    zload += an * sin(kn * x);
     
                    t0 = te * ((D * kn4 / drho_g) + 1.0);
                    Yt = 1.0 + ((te / t0) - 1.0) * exp(-t / t0);
                    bn = an * rhoc * grav * Yt / drho_g;
                    wfourier -= bn * sin(kn * x);
               }
     
               fprintf(out, "%g %g %g %g %g\n", x, zload, wfourier, x / 1000.0, wfourier / 1000.0);
          }
     }
     
     fclose(out);
     printf("(data for viscoelastic deflection is written to: %s)\n", filename);
}


void write_viscous_deflection(
     void)
{
     int i, k, n;
     FILE *out = fopen(filename, "w");
     double dx, x, zload, wfourier, an, bn, kn, kn2, kn4;
     double t, t_ref, t0, lambda_c, F;

     write_params(out);

     F = F0 * F_factor;
     t_ref = 4.0 * Dv * drho_g / (F * F);

     if (t_hat_out_size <= 0)
     {
          t_hat_out[0] = 0.0;
          t_hat_out[1] = 0.001;
          t_hat_out[2] = 0.01;
          t_hat_out[3] = 0.1;
          t_hat_out[4] = 1.0;
          t_hat_out_size = 5;
     }

     dx = L / grid_size;
     lambda_c = 2.0 * M_PI * sqrt(F / drho_g);

     for (i = 0; i < t_hat_out_size; i++)
     {
          t = t_hat_out[i] * t_ref;

          fprintf(out, "\n");
          fprintf(out, "#( lambda_c=%g [m], L=%g [m], tu = %g [Ma], F = %g [N] )#\n", 
               lambda_c, L, (t_ref / SEC_IN_MA), F);
          fprintf(out, "#( t=%g [Ma], tHat=%g )#\n", (t / SEC_IN_MA), (t / t_ref));
          fprintf(out, "#( x:0 zLoad:1 wFourier:2 x:3 wFourier:4)#\n");

          for (x = 0.0; x < L; x += dx)
          {
               zload = 0.0;
               wfourier = 0.0;

               for (n = 1; n < infinity; n++)
               {
                    an = 0.0;

                    for (k = 0; k < load_size - 1; k++)
                         an += get_rectangle_fourier_part(n, k);

                    kn = n * pi / L;
                    kn2 = kn * kn;
                    kn4 = kn2 * kn2;

                    zload += an * sin(kn * x);

		    /* 
                    ** The bn are the Fourier coefficients of an 
		    ** elastic deflection from a rectangular load with
                    ** Fourier coefficients an. 
                    */

                    bn = an * rhoc * grav / (D * kn4 + drho_g);
                    t0 = Dv * kn4 / (F * kn2 - drho_g);
                    wfourier -= bn * exp(t / t0) * sin(kn * x);

                    if (debug_wave_lengths && (i == 0) && (x < 0.001))
                    {
                         double lambda_n = 2 * M_PI / kn;
                         if (n == 1) printf("Fourier coefficients at x=%g:\n", x);
                         printf("n=%2d, an=%8.1e, lambda=%6.1e, lambda_c=%6.1e, (%s)\n",
                              n, bn, lambda_n, lambda_c,
                              ((lambda_n > lambda_c) ? "stable" : "UNSTABLE"));
                    }
               }

               fprintf(out, "%g %g %g %g %g\n", x, zload, wfourier, x / 1000.0, wfourier / 1000.0);
          }
     }

     fclose(out);
     printf("(data for viscous deflection is written to: %s)\n", filename);
}


void write_params(
     FILE *out)
{
     fprintf(out, "%s", (out == stdout) ? "(" : "#( ");
     fprintf(out, "E=%g [Pa], Te=%g [m], nu=%g [-], (length) L=%g [m] ", E, Te, nu, L);
     fprintf(out, "rhoc=%g [kg/m3], rhom=%g [kg/m3], ", rhoc, rhom);
     fprintf(out, "mu=%g [Pas]", mu);
     fprintf(out, "%s\n", (out == stdout) ? ")" : " )#");
}


double get_rectangle_fourier_part(
     int n,
     int k)
{
     double x1 = xload[k];
     double x2 = xload[k + 1];
     double h1 = hload[k];
     double h2 = hload[k + 1];
     double I = 0;

     I -= get_rectangle_dI(n, x1, x1, x2, h1, h2);
     I += get_rectangle_dI(n, x2, x1, x2, h1, h2);

     I *= (2.0 / L);

     return I;
}


double get_rectangle_dI(
     int n,
     double x,
     double x1,
     double x2,
     double h1,
     double h2)
{
     double a12 = (h2 - h1) / (x2 - x1);
     double kn = n * pi / L;
     double kn2 = kn * kn;
     double II1 = (x / kn) * cos(kn * x) + (1 / kn2) * sin(kn * x);
     double II2 = - (h1 - a12 * x1) * cos(kn * x) / kn;

     double dI = (a12 * II1 - II2);

     return dI;
}


double get_point_part(
     int k,
     double x)
{
     double xm = (xload[k + 1] + xload[k]) / 2.0;
     double dx = (xload[k + 1] - xload[k]);
     double hm = (hload[k + 1] + hload[k]) / 2.0;
     double dA = dx * hm;
     double V0 = rhoc * dA * grav;
     double w = get_point_load_deflection(x, xm, V0);
     return w;
}


double get_point_load_deflection(
     double x,
     double xm,
     double V0)
{
     double wmax = V0 / (2.0 * drho_g * alpha);
     double xref = (x > xm) ? (x - xm) : (xm - x);
     double xhat = xref / alpha;
     double term = exp(- xhat) * (cos(xhat) + sin(xhat));
     double w = wmax * term;
     return w;
}


double get_one_point_load(
     double x)
{
     double area = get_total_area();
     double V0 = rhoc * area * grav;
     double xm = L / 2.0;
     double w = get_point_load_deflection(x, xm, V0);
     return w;
}


double get_total_area(
     void)
{
     int k;
     double dx, hm;
     double area = 0;

     for (k = 0; k < load_size - 1; k++)
     {
          dx = (xload[k + 1] - xload[k]);
          hm = (hload[k + 1] + hload[k]) / 2.0;
          area += dx * hm;
     }

     return area;
}



