/*
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; version 2 of the License.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   (c) Copyright 1998-2009 by M. Wangen, IFE, Box 40, N-2007 Kjeller, Norway.
**
**   File: vr.c
**   Title: Compute vitrinite reflectance for simple cases
**   Date: Version 1.0, April 1998
**   Author: Magnus Wangen, IFE, magnus@ife.no
**
**   $Id$
*/

/*
**   History:
**   Version 1.0, Apr 1998: First version for UNIK use.
*/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define R 8.314
#define MAX_STEPS 32
#define MAX_WORDS 256

#define REPORT_DAYS 1
#define REPORT_MILL_YEARS 2

#define METHOD_PIECEWISE_CONST 1
#define METHOD_LINEAR_APPROX   2

#define ABS(x) ((x) > 0.0 ? (x) : (-(x))) 
#define MAX(x, y) ((x) > (y) ? (x) : (y)) 
#define MIN(x, y) ((x) > (y) ? (y) : (x))
#define MATCH(a, b) (strcmp(a,b) == 0)

typedef struct {
     double zpos;
     double temp;
     double thickness;
     double t0;
} SillParams;

typedef struct {
     double burial_rate;
     double dTdz;
     double t0;
     double t1;
     double t2;
} BurialHistParams;

typedef void (*DO_TIME_INTERVAL_FUNC)(double dt_sec, double temp1, double temp2,
     double *tr_out, double *ro_out);

int main(int arc, char **argv);
void init_kinetics(void);
void read_options(int argc, char **argv);
void show_options(FILE *out);
void test1();
void run_experiment1(const char *filename);
void run_experiment2(const char *filename, double temp_max, 
     double time_interval_sec, int n_decades, int info_code);
void run_experiment3(const char *filename);
void run_experiment4(const char *filename);
void run_experiment_old(const char *filename);
void run_burial_hist_with_sill(FILE *out,
     SillParams *sill, BurialHistParams *hist);
void run_burial_hist_with_sill_v2(FILE *out,
     SillParams *sill, BurialHistParams *hist);
double get_temp_from_the_sill(double z, double t, SillParams *sill);
double get_analytical_temp(double zhat, double that, double depth_hat);
double get_ffunc(double zhat, double that);
void do_temp_interval(double temp1_input, double temp2_input, 
     double dt_input, double *tr, double *ro);
void do_temp_interval_simple(double temp1_input, double temp2_input, 
     double dt_input, double *tr, double *ro);
void do_one_time_step(double dt, double temp, 
     double *tr_out, double *ro_out);
void do_temp_interval_approx(double dt_sec, double temp1, double temp2, 
     double *tr_out, double *ro_out);

static int
     n_steps = 20,
     is_first_time = 1,
     debug_interval_func = 0,
     num_method = METHOD_PIECEWISE_CONST;

static double
     A[MAX_STEPS],
     E[MAX_STEPS],
     x0[MAX_STEPS],
     x1[MAX_STEPS];

static double
     kappa = 1.0e-6;

static DO_TIME_INTERVAL_FUNC 
     do_temp_interval_ptr = do_temp_interval_simple;


int main(
     int argc, 
     char **argv)
{
     double sec_in_day = 60.0 * 60.0 * 24.0;
     double sec_in_year = sec_in_day * 365.0;
     double sec_in_1000_years = sec_in_year * 1000.0;
     double sec_in_million_years = sec_in_year * 1.0e+6;

     read_options(argc, argv);

     run_experiment1("dataEasyRoTemp.xy");
     run_experiment2("dataEasyRoRate1.xy", 250.0, sec_in_million_years, 4, REPORT_MILL_YEARS);
     run_experiment2("dataEasyRoRate2.xy", 500.0, sec_in_day, 4, REPORT_DAYS);
     run_experiment2("dataSB2a2.xy", 600.0, 120.0 * sec_in_million_years, 1, REPORT_DAYS);
     run_experiment2("dataSB2b2.xy", 600.0, 600.0 * sec_in_1000_years, 1, REPORT_DAYS);
     run_experiment2("dataSB2c2.xy", 600.0, 60.0 * sec_in_year, 1, REPORT_DAYS);
     run_experiment3("dataSill100m.xy");
     run_experiment4("dataSillThickness.xy");

     return 0;
}


void read_options(
     int argc,
     char **argv)
{
     int i;
     int got_errors = 0;

     for (i = 1; i < argc; i++)
          if      (MATCH(argv[i], "-h")) {show_options(stdout); exit(1);}
          else if (MATCH(argv[i], "-debug"))  debug_interval_func = 1;
          else if (MATCH(argv[i], "-simple")) do_temp_interval_ptr = do_temp_interval_simple;
          else if (MATCH(argv[i], "-approx")) do_temp_interval_ptr = do_temp_interval_approx;
          else
          {
               fprintf(stderr, "Unknown option: %s\n", argv[i]);
               got_errors++;
          }

     if (got_errors)
          exit(1);
}


void show_options(
     FILE *out)
{
     fprintf(out, "Options:\n");
     fprintf(out, "     -simple    Simple constant step heating.\n");
     fprintf(out, "     -approx    Approximate linear heating.\n");
     fprintf(out, "     -debug     Report methods being used.\n");
     fprintf(out, "     -h         Print this message.\n");
}


void init_kinetics(
     void)
{
     int i;

     /* Sweeney and Burnham 1990 E is in kcal/mole. */
     A[0] = 1e+13; E[0] = 34; x0[0] = 0.03;
     A[1] = 1e+13; E[1] = 36; x0[1] = 0.03;
     A[2] = 1e+13; E[2] = 38; x0[2] = 0.04;
     A[3] = 1e+13; E[3] = 40; x0[3] = 0.04;
     A[4] = 1e+13; E[4] = 42; x0[4] = 0.05;
     A[5] = 1e+13; E[5] = 44; x0[5] = 0.05;
     A[6] = 1e+13; E[6] = 46; x0[6] = 0.06;
     A[7] = 1e+13; E[7] = 48; x0[7] = 0.04;
     A[8] = 1e+13; E[8] = 50; x0[8] = 0.04;
     A[9] = 1e+13; E[9] = 52; x0[9] = 0.07;
     A[10] = 1e+13; E[10] = 54; x0[10] = 0.06;
     A[11] = 1e+13; E[11] = 56; x0[11] = 0.06;
     A[12] = 1e+13; E[12] = 58; x0[12] = 0.06;
     A[13] = 1e+13; E[13] = 60; x0[13] = 0.05;
     A[14] = 1e+13; E[14] = 62; x0[14] = 0.05;
     A[15] = 1e+13; E[15] = 64; x0[15] = 0.04;
     A[16] = 1e+13; E[16] = 66; x0[16] = 0.03;
     A[17] = 1e+13; E[17] = 68; x0[17] = 0.02;
     A[18] = 1e+13; E[18] = 70; x0[18] = 0.02;
     A[19] = 1e+13; E[19] = 72; x0[19] = 0.01;

     n_steps = 20;
     is_first_time = debug_interval_func;

     for (i = 0; i < n_steps; i++)
          x1[i] = x0[i];
}


void test1(
     void)
{
     double sec_in_years = 60.0 * 60.0 * 24.0 * 365.0;
     double temp = 250.0;
     double tr, ro, t_sec, dt_sec, t_years, log10_t_years, t_million_years;

     init_kinetics();
     fprintf(stdout, "#( temp = %g )#\n", temp);

     dt_sec = 1.0;
     t_sec = 0.0;
     ro = 0.0;

     while(ro < 8.0)
     {
          do_one_time_step(dt_sec, temp, &tr, &ro);
          t_sec += dt_sec;
          dt_sec *= 2.0;

          t_years = t_sec / sec_in_years;
          log10_t_years = log10(t_years);
          t_million_years = t_years / 1.0e+6;

          fprintf(stdout, "%g %g %g %g\n", t_million_years, log10_t_years, tr, ro);
     }
}


void run_experiment1(
     const char *filename)
{
     FILE *out = fopen(filename, "w");
     double sec_in_years = 60.0 * 60.0 * 24.0 * 365.0;
     double temp, tr, ro, t_sec, dt_sec, t_years, log10_t_years;

     if (out == NULL)
     {
          fprintf(stdout, "Can't open file: %s\n", filename);
          exit(1);
     }

     for (temp = 50.0; temp < 351.0; temp += 50)
     {
          init_kinetics();
          fprintf(out, "#( temp = %g )#\n", temp);

          dt_sec = 1.0e+4;
          t_sec = 0.0;
          ro = 0.0;

          while(ro < 4.5)
          {
               do_one_time_step(dt_sec, temp, &tr, &ro);
               t_sec += dt_sec;
               dt_sec *= 2.0;

               t_years = t_sec / sec_in_years;
               log10_t_years = log10(t_years);

               fprintf(out, "%g %g %g\n", log10_t_years, tr, ro);
          }
     }

     fclose(out);
     printf("isothermal VR is written to: %s\n", filename);
}


void run_experiment2(
     const char *filename,
     double temp_max,
     double time_interval_sec,
     int n_decades,
     int info_code)
{
     FILE *out = fopen(filename, "w");
     double sec_in_day = 60.0 * 60.0 * 24.0;
     double sec_in_million_years = 60.0 * 60.0 * 24.0 * 365.0 * 1.0e+6;
     int steps = 100;
     int i, n;

     if (out == NULL)
     {
          fprintf(stdout, "Can't open  %s\n", filename);
          exit(1);
     }

     for (n = 0; n < n_decades; n++)
     {
          double dt_sec = time_interval_sec / steps;
          double dt_million_years = dt_sec / sec_in_million_years;
          double dtemp = temp_max / steps;
          double tr, ro;

          fprintf(out, "\n");

          if (info_code & REPORT_DAYS)
          {
               double interval_days = time_interval_sec / sec_in_day;
               printf("(interval length = %g [days])\n", interval_days);
               fprintf(out, "#( interval length = %g [days] )#\n", interval_days);
          }

          if (info_code & REPORT_MILL_YEARS)
          {
               double interval_mill_years = time_interval_sec / sec_in_million_years;
               printf("(interval length = %g [Ma])\n", interval_mill_years);
               fprintf(out, "#( interval length = %g [Ma] )#\n", interval_mill_years);
          }

          fprintf(out, "#( %16s %12s %12s %12s %12s )#\n", 
               "log10(t)", "tr", "ro", "log10(ro)", "temp");

          fprintf(out, "#( %16s %12s %12s %12s %12s )#\n", 
               "[log10(years)]", "[-]", "[-]", "[-]", "[C]");

          init_kinetics();
          ro = 0;

          for (i = 1; (i <= steps) && (ro < 4.5); i++)
          {
               double log10_t_years = log10(i * dt_million_years) + 6.0;
               double temp0 = (i - 1) * dtemp;
               double temp1 = temp0 + dtemp;

               /***
               do_temp_interval_simple(temp0, temp1, dt_sec, &tr, &ro);
               do_temp_interval_approx(temp0, temp1, dt_sec, &tr, &ro);
               ***/
               do_temp_interval_ptr(temp0, temp1, dt_sec, &tr, &ro);

               fprintf(out, "   %16g %12g %12g %12g %12g\n", 
                    log10_t_years, tr, ro, log10(ro), temp1);
          }

          time_interval_sec *= 10.0;
     }

     fclose(out);
     printf("VR for const heating rate is written to: %s\n", filename);
}


void run_experiment_old(
     const char *filename)
{
     double sec_in_mill_years = 60 * 60 * 24 * 365 * 10e+6;
     double dTdz = 35.0e-3;
     double t0 = 0.0 * sec_in_mill_years;
     double t1 = 200.0 * sec_in_mill_years;
     double t2 = 250.0 * sec_in_mill_years;
     double dt0 = t1 / 500.0;
     double burial_rate = 6000.0 / t2;
     double sill_temp[10] = {0.0, 500.0, 750.0, 1000.0, 1250.0, -1.0};

     double dz_after_sill = burial_rate * (t2 - t1);
     FILE *out = fopen(filename, "w");

     double temp0, temp1, temp2, tempA, tempB, temp_ref;
     double sill_zpos_obs, t0_sill, t_sill_max, dt_sill_first;
     double z, t, dt, tr, ro; 
     SillParams sill;
     int i;

     if (out == NULL)
     {
          fprintf(stdout, "Can't open  %s\n", filename);
          exit(1);
     }

     temp_ref = dTdz * burial_rate * (t1 - t0);
     sill_zpos_obs = 3860.0;
     sill.temp = 1000.0 - temp_ref;
     sill.zpos = sill_zpos_obs - dz_after_sill;
     sill.thickness = 92.0;

     t0_sill = sill.thickness * sill.thickness / kappa;
     t_sill_max = 10 * t0_sill;
     dt_sill_first = 0.01 * t0_sill;

     printf("(===== SILL =====)\n");
     printf("(burial after sill emplacement: %g [m])\n", dz_after_sill);
     printf("(temp at depth of sill emplacement: %g [C])\n", temp_ref);
     printf("(sill zpos1 = %g)\n", sill_zpos_obs - 0.5 * sill.thickness);
     printf("(sill zpos2 = %g)\n", sill_zpos_obs + 0.5 * sill.thickness);
     printf("(sill time span = %g [Ma])\n", t_sill_max / sec_in_mill_years);

     for (i = 0; sill_temp[i] >= 0.0; i++)
     {
          fprintf(out, "\n");
          fprintf(out, "#( sill-temp=%g [C] )#\n", sill_temp[i]);
          printf("(temp = %g [C])\n", sill_temp[i]);
          sill.temp = sill_temp[i] - temp_ref;

          for (t0 = 0.0; t0 < t1; t0 += dt0)
          {
               init_kinetics();
     
               /* (1) Burial until t1 */
               temp0 = 0.0;
               temp1 = dTdz * burial_rate * (t1 - t0);
               do_temp_interval_ptr(temp0, temp1, (t1 - t0), &tr, &ro);

               /* (2) Thermal perturbation by sill. */
               z = burial_rate * (t1 - t0);
               tempA = temp1;
               tempB = temp1 + get_temp_from_the_sill(z, 0.0, &sill);
               dt = dt_sill_first;

               for (t = dt_sill_first; t < t_sill_max; )
               {
                    tempA = tempB;
                    tempB = temp1 + get_temp_from_the_sill(z, t, &sill);
                    do_temp_interval_ptr(tempA, tempB, dt, &tr, &ro);
     
                    dt *= 1.5;
                    t += dt;
               }

               /* (3) Burial overtime span t1 to t2. */
               temp1 = dTdz * burial_rate * (t1 - t0);
               temp2 = dTdz * burial_rate * (t2 - t0);
               do_temp_interval_ptr(temp1, temp2, (t2 - t1), &tr, &ro);
     
               z = burial_rate * (t2 - t0);
               fprintf(out, "%g %g %g\n", -z, tr, ro);
          }
     }

     fclose(out);
     printf("VR for sills is written to: %s\n", filename);
}


void run_experiment3(
     const char *filename)
{
     SillParams sill;
     BurialHistParams hist;
     FILE *out = fopen(filename, "w");
     double sec_in_mill_years = 60 * 60 * 24 * 365 * 10e+6;
     double sill_temp[10] = {0.0, 500.0, 750.0, 1000.0, 1250.0, -1.0};
     double temp_ref, sill_zpos_obs, dz_after_sill;
     int i;

     if (out == NULL)
     {
          fprintf(stdout, "Can't open  %s\n", filename);
          exit(1);
     }

     fprintf(out, "#( z:0   tr:1  ro:2  )#\n");
     fprintf(out, "#( [-]:0 [-]:1 [-]:2 )#\n");

     hist.dTdz = 35.0e-3;
     hist.t0 = 0.0 * sec_in_mill_years;
     hist.t1 = 200.0 * sec_in_mill_years;
     hist.t2 = 250.0 * sec_in_mill_years;
     hist.burial_rate = 6000.0 / hist.t2;

     dz_after_sill = hist.burial_rate * (hist.t2 - hist.t1);
     sill_zpos_obs = 3860.0;
     temp_ref = hist.dTdz * (sill_zpos_obs - dz_after_sill);
     sill.temp = 1000.0 - temp_ref;
     sill.zpos = sill_zpos_obs - dz_after_sill;
     sill.thickness = 92.0;
     sill.t0 = sill.thickness * sill.thickness / kappa;

     printf("(===== SILL =====)\n");
     printf("(burial after sill emplacement: %g [m])\n", dz_after_sill);
     printf("(temp at depth of sill emplacement: %g [C])\n", temp_ref);
     printf("(sill zpos1 = %g)\n", sill_zpos_obs - 0.5 * sill.thickness);
     printf("(sill zpos2 = %g)\n", sill_zpos_obs + 0.5 * sill.thickness);
     printf("(sill timp span = %g [Ma])\n", 10.0 * sill.t0 / sec_in_mill_years);

     for (i = 0; sill_temp[i] >= 0.0; i++)
     {
          fprintf(out, "\n");
          fprintf(out, "#( sill-temp=%g [C] )#\n", sill_temp[i]);
          printf("(temp = %g [C])\n", sill_temp[i]);
          sill.temp = sill_temp[i] - temp_ref;
          run_burial_hist_with_sill(out, &sill, &hist);
     }

     fclose(out);
     printf("VR for sills is written to: %s\n", filename);
}


void run_experiment4(
     const char *filename)
{
     SillParams sill;
     BurialHistParams hist;
     FILE *out = fopen(filename, "w");
     double sec_in_mill_years = 60 * 60 * 24 * 365 * 10e+6;
     double thickness[10] = {10.0, 100.0, 500.0, 1000.0, -1.0};
     int i;

     if (out == NULL)
     {
          fprintf(stdout, "Can't open  %s\n", filename);
          exit(1);
     }

     fprintf(out, "#( z:0   tr:1  ro:2  )#\n");
     fprintf(out, "#( [-]:0 [-]:1 [-]:2 )#\n");

     hist.dTdz = 35.0e-3;
     hist.t0 = 0.0 * sec_in_mill_years;
     hist.t1 = 200.0 * sec_in_mill_years;
     hist.t2 = 250.0 * sec_in_mill_years;
     hist.burial_rate = 6000.0 / hist.t2;

     for (i = 0; thickness[i] > 0.0; i++)
     {
          double dz_after_sill = hist.burial_rate * (hist.t2 - hist.t1);
          double sill_zpos_obs = 4000.0;
          double temp_ref = hist.dTdz * (sill_zpos_obs - dz_after_sill);

          sill.temp = 1000.0 - temp_ref;
          sill.zpos = sill_zpos_obs - dz_after_sill;
          sill.thickness = thickness[i];
          sill.t0 = sill.thickness * sill.thickness / kappa;

          printf("(===== SILL =====)\n");
          printf("(burial after sill emplacement: %g [m])\n", dz_after_sill);
          printf("(temp at depth of sill emplacement: %g [C])\n", temp_ref);
          printf("(sill zpos1 = %g)\n", sill_zpos_obs - 0.5 * sill.thickness);
          printf("(sill zpos2 = %g)\n", sill_zpos_obs + 0.5 * sill.thickness);
          printf("(sill time span = %g [Ma])\n", 10.0 * sill.t0 / sec_in_mill_years);

          fprintf(out, "\n");
          fprintf(out, "#( sill-thickness=%g [m] )#\n", sill.thickness);
          run_burial_hist_with_sill_v2(out, &sill, &hist);
     }

     fclose(out);
     printf("VR for sills is written to: %s\n", filename);
}


void run_burial_hist_with_sill(
     FILE *out,
     SillParams *sill,
     BurialHistParams *hist)
{
     double t0 = hist->t0;
     double t1 = hist->t1;
     double t2 = hist->t2;
     double dt0 = t1 / 500.0;
     double dTdz = hist->dTdz;
     double burial_rate = hist->burial_rate;
     double dt_sill_first = 0.001 * sill->t0;
     double t_sill_max = 1.0e+5 * sill->t0;
     double temp0, temp1, temp2, tempA, tempB;
     double z, t, dt, tr, ro;

     for (t0 = 0.0; t0 < t1; t0 += dt0)
     {
          init_kinetics();

          /* (1) Burial until t1 */
          temp0 = 0.0;
          temp1 = dTdz * burial_rate * (t1 - t0);
          do_temp_interval_ptr(temp0, temp1, (t1 - t0), &tr, &ro);

          /* (2) Thermal perturbation by sill. */
          z = burial_rate * (t1 - t0);
          tempA = temp1;
          tempB = temp1 + get_temp_from_the_sill(z, 0.0, sill);
          dt = dt_sill_first;

          for (t = dt_sill_first; t < t_sill_max; )
          {
               tempA = tempB;
               tempB = temp1 + get_temp_from_the_sill(z, t, sill);
               do_temp_interval_ptr(tempA, tempB, dt, &tr, &ro);

               dt *= 1.5;
               t += dt;
          }

          /* (3) Burial overtime span t1 to t2. */
          temp1 = dTdz * burial_rate * (t1 - t0);
          temp2 = dTdz * burial_rate * (t2 - t0);
          do_temp_interval_ptr(temp1, temp2, (t2 - t1), &tr, &ro);

          z = burial_rate * (t2 - t0);
          fprintf(out, "%g %g %g\n", -z, tr, ro);
     }
}


void run_burial_hist_with_sill_v2(
     FILE *out,
     SillParams *sill,
     BurialHistParams *hist)
{
     double t0 = hist->t0;
     double t1 = hist->t1;
     double t2 = hist->t2;
     double dt0 = t1 / 500.0;
     double dTdz = hist->dTdz;
     double burial_rate = hist->burial_rate;
     double dt_sill_first = 0.001 * sill->t0;
     double t_sill_max = 1.0+5 * sill->t0;
     double temp0, temp1, temp2, tempA, tempB;
     double z, t, dt, tr, ro;

     for (t0 = 0.0; t0 < t1; t0 += dt0)
     {
          init_kinetics();

          /* (2) Thermal perturbation by sill. */
          z = burial_rate * (t1 - t0);
          tempA = temp1;
          tempB = temp1 + get_temp_from_the_sill(z, 0.0, sill);
          dt = dt_sill_first;

          for (t = dt_sill_first; t < t_sill_max; )
          {
               tempA = tempB;
               tempB = temp1 + get_temp_from_the_sill(z, t, sill);
               do_temp_interval_ptr(tempA, tempB, dt, &tr, &ro);

               dt *= 1.5;
               t += dt;
          }

          /* (3) Burial overtime span t1 to t2. */
          temp1 = dTdz * burial_rate * (t1 - t0);
          temp2 = dTdz * burial_rate * (t2 - t0);
          do_temp_interval_ptr(temp1, temp2, (t2 - t1), &tr, &ro);

          /* (1) Burial until t1 */
          temp0 = 0.0;
          temp1 = dTdz * burial_rate * (t1 - t0);
          do_temp_interval_ptr(temp0, temp1, (t1 - t0), &tr, &ro);

          z = burial_rate * (t2 - t0);
          fprintf(out, "%g %g %g\n", -z, tr, ro);
     }
}


double get_temp_from_the_sill(
     double z,
     double t,
     SillParams *sill)
{
     double zpos = sill->zpos;
     double temp_ref = sill->temp;
     double href = sill->thickness / 2.0;

     double zhat = (z - zpos) / href;
     double t0 = href * href / kappa;
     double that = t / t0;
     double depth_hat = zpos / href;
     double temp = temp_ref * get_analytical_temp(zhat, that, depth_hat);
     return temp;
}


double get_analytical_temp(
     double zhat,
     double that,
     double depth_hat)
{
     double term1, term2;

     if (ABS(zhat - 1.0) < 0.001)
          zhat =  0.999;
     else if (ABS(zhat + 1.0) < 0.001)
          zhat = -0.999;

     term1 = get_ffunc(zhat, that);
     term2 = get_ffunc(2.0 * depth_hat - zhat, that);

     return (term1 - term2);
}


double get_ffunc(
     double zhat,
     double that)
{
     double sqrt_that = sqrt(that);
     double term1 = 0.5 * (zhat + 1.0) / sqrt_that;
     double term2 = 0.5 * (zhat - 1.0) / sqrt_that;
     double term3 = erf(term1);
     double term4 = erf(term2);
     double ff = 0.5 * (term3 - term4);

     return ff;
}


void do_temp_interval(
     double temp1_input,
     double temp2_input,
     double dt_input_sec,
     double *tr,
     double *ro)
{
     switch (num_method)
     {
          case METHOD_PIECEWISE_CONST: 
               do_temp_interval_simple(temp1_input, temp2_input, dt_input_sec, tr, ro);
               break;

          case METHOD_LINEAR_APPROX:
               do_temp_interval_approx(temp1_input, temp2_input, dt_input_sec, tr, ro);
               break;

          default:
               fprintf(stderr, "Unknown numerical method!\n");
               exit(1);
     }
}


void do_temp_interval_simple(
     double temp1_input,
     double temp2_input,
     double dt_input_sec,
     double *tr,
     double *ro)
{
     double dtemp_min = 0.5;
     double dtemp_input = temp2_input - temp1_input;
     int min_steps = (int) (ABS(dtemp_input) / dtemp_min);
     int steps = MAX(min_steps, 1);
     double dtemp = dtemp_input / steps;
     double dt_sec = dt_input_sec / steps;
     double temp;
     int i;

     for (i = 0; i < steps; i++)
     {
          temp = temp1_input + (i + 0.5)  * dtemp;
          do_one_time_step(dt_sec, temp, tr, ro);
     }
}


void do_one_time_step(
     double dt_sec, 
     double temp,
     double *tr_out,
     double *ro_out)
{
     double RT, E2, log_k_dt, k_dt, factor, tr, ro;
     double sum0 = 0.0;
     double sum1 = 0.0;
     int i;

     if (is_first_time)
     {
          printf("(is using simple-function)\n");
          is_first_time = 0;
     }

     for (i = 0; i < n_steps; i++)
     {
          RT = R * (273.0 + temp);
          E2 = 1000 * 4.1868 * E[i];  /* [kcal/mole] to [joule/mole] */
          log_k_dt = log(A[i] * dt_sec) - (E2 / RT);
          k_dt = exp(log_k_dt);
          factor = exp(-k_dt);
          x1[i] *= factor;

          sum0 += x0[i];
          sum1 += x1[i];
     }

     tr = (sum0 - sum1);
     ro = exp(-1.6 + 3.7 * (tr));

     *tr_out = tr;
     *ro_out = ro;
}


void do_temp_interval_approx(
     double temp1,
     double temp2,
     double dt_sec,
     double *tr_out,
     double *ro_out)
{
     double E2, Tf, N, u1, u2, F1, F2, factor, tr, ro;
     double heating_rate = (temp2 - temp1) / dt_sec;
     double sum0 = 0.0;
     double sum1 = 0.0;
     int i;

     if (is_first_time)
     {
          printf("(is using approx-function)\n");
          is_first_time = 0;
     }

     for (i = 0; i < n_steps; i++)
     {
          E2 = 1000 * 4.1868 * E[i];  /* [kcal/mole] to [joule/mole] */
          Tf = E2 / R;
          N = A[i] * Tf / heating_rate;
          u1 = (temp1 + 273.0) / Tf;
          u2 = (temp2 + 273.0) / Tf;

          F1 = u1 * u1 * exp(-1.0 / u1);
          F2 = u2 * u2 * exp(-1.0 / u2);
          factor = exp(- N * (F2 - F1));
          x1[i] *= factor;

          sum0 += x0[i];
          sum1 += x1[i];
     }

     tr = (sum0 - sum1);
     ro = exp(-1.6 + 3.7 * (tr));

     *tr_out = tr;
     *ro_out = ro;
}


